{- OPTIONS_GHC -fbang-patterns -O2 #-}
{-
 Module      : Data.ByteString.KMP
 Copyright   : (c) Justin Bailey 2007 (jgbailey@gmail.com)
               (c) Chris Kuklewicz 2007 (haskell@list.mightyreason.com)
               (c) Daniel Fischer  (daniel.is.fischer@web.de)
 License     : BSD-style

 Maintainer  : jgbailey@gmail.com
 Stability   : experimental

 Implements Knuth-Morris-Pratt fast string matching for Strict and Lazy
 ByteStrings. Original code by Justin Bailey, with significant
 enhancemants and optimizations by Chris Kuklewicz. Further enhancements to
 the computeLookup function provided by Daniel Fischer.

 Tests can be run by loading the module and evaluating "allTests".
-}
module Data.ByteString.KMP (kmpMatchLL, kmpMatchLS, kmpMatchSS, kmpMatchSL, kmpMatchSSs)

where

import qualified Data.Array.Base as Base (unsafeAt, unsafeWrite, unsafeRead)
import qualified Data.Array.Unboxed as Unboxed (UArray)
import qualified Data.Array.IArray as IArray (array)
import qualified Data.Array.ST as ST (runSTUArray, newArray_)
import qualified Data.List as List (map, filter, length, null, take, maximum, foldr, all, drop)
import qualified Data.ByteString.Lazy as L
import qualified Data.ByteString as S
import qualified Data.ByteString.Base as B (unsafeHead,unsafeTail,unsafeDrop,unsafeIndex)
import GHC.Int (Int64)

import Test.QuickCheck
import Debug.Trace (trace)

{-|
  Returns list of indices for a given substring in a search string, or the empty
  list if none were found.
  
  Uses the Knuth-Morris-Pratt fast string matching algorithm.
-}  
{-# INLINE kmpMatchLL #-}
kmpMatchLL :: L.ByteString -- ^ Pattern to search for.
  -> L.ByteString -- ^ String to search.
  -> [Int64] -- ^ List of indices where string was found.
kmpMatchLL pat = let search = kmpMatchSSs (S.concat (L.toChunks pat)) in search . L.toChunks

{-|
  Returns list of indices for a given substring in a search string, or the empty
  list if none were found.
  
  Uses the Knuth-Morris-Pratt fast string matching algorithm.
-}  
{-# INLINE kmpMatchLS #-}
kmpMatchLS :: L.ByteString -- ^ Pattern to search for.
  -> S.ByteString -- ^ String to search.
  -> [Int64] -- ^ List of indices where string was found.
kmpMatchLS pat = let search = kmpMatchSSs (S.concat (L.toChunks pat)) in search . (:[])

{-|
  Returns list of indices for a given substring in a search string, or the empty
  list if none were found.
  
  Uses the Knuth-Morris-Pratt fast string matching algorithm.
-}  
{-# INLINE kmpMatchSS #-}
kmpMatchSS :: S.ByteString -- ^ Pattern to search for.
  -> S.ByteString -- ^ String to search.
  -> [Int64] -- ^ List of indices where string was found.
kmpMatchSS pat = let search = kmpMatchSSs pat in search . (:[])

{-|
  Returns list of indices for a given substring in a search string, or the empty
  list if none were found.
  
  Uses the Knuth-Morris-Pratt fast string matching algorithm.
-}  
{-# INLINE kmpMatchSL #-}
kmpMatchSL :: S.ByteString -- ^ Pattern to search for.
  -> L.ByteString -- ^ String to search.
  -> [Int64] -- ^ List of indices where string was found.
kmpMatchSL pat = let search = kmpMatchSSs pat in search . L.toChunks

kmpMatchSSs :: S.ByteString -> [S.ByteString] -> [Int64]
kmpMatchSSs pat =
  let !patLen = S.length pat -- Evaluate S.length once; 
      !lookupTable = {-# SCC "kmpMatchSSs_computeLookup'" #-} computeLookup pat -- lower bound of UArray must be 0 for Base.unsafeAt, but index 0 will never be looked up
      searcher :: Int64 -> Int -> [S.ByteString] -> [Int64]
      searcher _ _ [] = []
      searcher !prior !patStart (!str:strRest) =
        let !strLen = S.length str -- Evaluate S.length once; 
            findMatch :: Int -> Int -> [Int64]
            findMatch !strIndex !patIndex | patIndex == patLen = (prior + fromIntegral strIndex - fromIntegral patLen) : findMatch strIndex 0 -- Success
                                        | strIndex == strLen = searcher (prior + fromIntegral strLen) patIndex strRest
                                        | otherwise =
              if (B.unsafeIndex str strIndex) == (B.unsafeIndex pat patIndex)
                then findMatch (succ strIndex) (succ patIndex)
                else if patIndex == 0
                       then findMatch (succ strIndex) 0
                       else findMatch strIndex (Base.unsafeAt lookupTable patIndex) -- here 1 <= patIndex <= patLen-1
        in
          findMatch 0 patStart
  in
    if patLen == 0
      then const [0]
      else searcher 0 0

{- |
 Given a string representing a search pattern, this function
 returns a function which represents, for each prefix of that
 pattern, the maximally long prefix of the pattern which is a suffix
 of the indicated pattern segment.

 If there is no such prefix, 0 is returned. For the very first
 element, -1 is always returned.
 -}
computeLookup :: S.ByteString -> Unboxed.UArray Int Int
computeLookup pat
   = {-# SCC "computeLookup_runSTUArray" #-} ST.runSTUArray
   (do ar <- ST.newArray_ (0,patLen)
       Base.unsafeWrite ar 0 (-1)
       Base.unsafeWrite ar 1 0
       let fill i
               | i > patLen = return ar
               | otherwise  =
                do let w = B.unsafeIndex pat (i-1)
                       help k
                           | k < 0 || w == B.unsafeIndex pat k =
                                do Base.unsafeWrite ar i (succ k)
                                   fill (i+1)
                           | otherwise = do m <- Base.unsafeRead ar k
                                            help m
                   n <- Base.unsafeRead ar (i-1)
                   help n
       {-# SCC "computeLookup_fill" #-} fill 2)
   where
     !patLen = S.length pat
     
-- Types, instances and utility functions for testing purposes.

newtype PatternChar = PatternChar Char
 deriving Show

instance Arbitrary PatternChar where
 arbitrary = oneof (List.map (return . PatternChar) ['a', 'b', 'c', 'd'])

-- Holds the search pattern, the search string, and the expected
-- position. Used to test for strings that have failed in the past.
data Regressions = R String String Int64
  deriving Show
  
instance Arbitrary Regressions where
  arbitrary =
    oneof $ map return 
      [R "ccb" "abcdcccb" 5
       , R "bbaa" "bdcdbbbbaa" 6
       , R "cccadadc" "adaccccadadc" 4
       , R "bbdbbdaabdb" "dbbdbbbdbbdaabdb" 5
       , R "bbdbbd" "dbbdbbbdbbd" 5
       , R "ccbb" "acccbb" 2
       , R "bbbb" "dddcaaaddaaabdacbcccabbada" (-1) -- This string has caused an infinite loop.
       , R "bbc" "bbbbc" 2
       , R "a" "" (-1)
       , R "" "a" 0
       , R "" "" 0
       ]
    
  
patternsToString :: [PatternChar] -> L.ByteString
patternsToString chars = L.pack $ List.foldr (\(PatternChar char) str -> (toEnum $ fromEnum char) : str) [] chars

patternsToStrictString :: [PatternChar] -> S.ByteString
patternsToStrictString chars = S.pack $ List.foldr (\(PatternChar char) str -> (toEnum $ fromEnum char) : str) [] chars

regressionsToLazy :: Regressions -> (L.ByteString, L.ByteString, Int64)
regressionsToLazy (R pat srch idx) = (toLazyBS pat, toLazyBS srch, idx)

toLazyBS :: String -> L.ByteString
toLazyBS = L.pack . List.map (toEnum . fromEnum)

toStrictBS :: String -> S.ByteString
toStrictBS = S.pack . List.map (toEnum . fromEnum)

-- Test that 0 and 1 element always return 0, if present.
prop_testZero :: [PatternChar] -> Property
prop_testZero pat =
 let table = computeLookup (patternsToStrictString pat)
 in
   trivial (List.null pat) $
   if List.length pat > 1
    then Base.unsafeAt table 0 == -1 && Base.unsafeAt table 1 == 0
    else Base.unsafeAt table 0 == -1

-- Test that all overlaps found are actually prefixes of the
-- pattern string
prop_testSubset :: [PatternChar] -> Property
prop_testSubset pat =
 let patStr = patternsToString pat
     table = computeLookup (patternsToStrictString pat)
     prefix len = L.take (fromIntegral len)
     testPrefix len =
       if Base.unsafeAt table len == 0
         then True
         else (prefix (Base.unsafeAt table len) patStr) `L.isPrefixOf` (prefix len patStr)
 in
   trivial (List.null pat) $
   List.all testPrefix [0 .. List.length pat - 1]


-- Test that the prefix given is the maximal prefix. That is,
-- add one more character makes it either equal to the string
-- or not a prefix.
prop_testCorrectPrefix :: [PatternChar] -> Property
prop_testCorrectPrefix pat =
 let patStr = patternsToString pat
     table = computeLookup (patternsToStrictString pat)
     isNeverSuffix len =
       let origPrefix = prefix len patStr
           -- Drop 1 to remove empty list
           allPrefixes = List.drop 1 $ L.inits origPrefix
       in
         List.all (\p -> L.null p || p == origPrefix || not ((L.reverse p) `L.isPrefixOf` (L.reverse origPrefix))) allPrefixes
     prefix len = L.take (fromIntegral len)
     -- True if the prefix returned from table for the length given produces
     -- a string which is a suffix of the original prefix.
     isRealSuffix len = (L.reverse (prefix (Base.unsafeAt table len) patStr)) `L.isPrefixOf` (L.reverse $ prefix len patStr)
     isLongestSuffix len =
       let prefixPlus = prefix (Base.unsafeAt table len + 1) patStr
           inputPrefix =  prefix len patStr
       in
         prefixPlus == inputPrefix ||
           not ((L.reverse prefixPlus) `L.isPrefixOf` (L.reverse inputPrefix))
     testTable len =
       if Base.unsafeAt table len == 0
         then isNeverSuffix len
         else isRealSuffix len &&
               isLongestSuffix len
 in
   trivial (List.null pat) $
   List.all testTable [0 .. List.length pat - 1]

-- Verify that, if a match is found, it is where it's supposed to be in
-- the string and it can be independently found by other means.
prop_testMatch :: [PatternChar] -> [PatternChar] -> Property
prop_testMatch pat search =
 let patStr = patternsToString pat
     searchStr = patternsToString search
     strictStr = patternsToStrictString search
     strictPat = patternsToStrictString pat
     patLen = L.length patStr
     searchLen = L.length searchStr
     matches = kmpMatchLL patStr searchStr
     verify matchIdx =
       if matchIdx > -1
         then (L.take patLen $ L.drop matchIdx $ searchStr) == patStr && strictPat `S.isSubstringOf` strictStr
         else (L.null patStr && L.null searchStr) || not (strictPat `S.isSubstringOf` strictStr)
 in
   trivial (List.null pat || L.null searchStr) $
   classify (patLen > searchLen) "Bigger pattern than search" $
   classify (patLen < searchLen) "Smaller pattern than search" $
   classify (patLen == searchLen) "Equal pattern and search" $
   classify (not (null matches)) "Match Exists" $
   classify (null matches) "Match Doesn't Exist" $
   all verify matches

-- Test a pattern that was known to fail.
prop_testBadPat :: [PatternChar] -> Property
prop_testBadPat search =
 let patStr = toLazyBS "bbc"
     patLen = L.length patStr
     searchStr = patternsToString search
     matches = kmpMatchLL patStr searchStr
     strictStr = patternsToStrictString search
     strictPat = toStrictBS "bbc"
     verify matchIdx =
            if matchIdx > -1
              then (L.take patLen $ L.drop matchIdx $ searchStr) == patStr && strictPat `S.isSubstringOf` strictStr
              else (L.null patStr && L.null searchStr) || not (strictPat `S.isSubstringOf` strictStr)
 in
   trivial (List.null search) $
     all verify matches

-- Test that a pattern on the end of the string is found OK.
prop_testEndPattern :: [PatternChar] -> [PatternChar] -> Property
prop_testEndPattern pat search =
 let patStr = patternsToString pat
     searchStr = patternsToString (search ++ pat)
     matches = kmpMatchLL patStr searchStr
     strictStr = patternsToStrictString (search ++ pat)
     strictPat = patternsToStrictString pat
     patLen = L.length patStr
     verify matchIdx = (L.take patLen $ L.drop matchIdx $ searchStr) == patStr && strictPat `S.isSubstringOf` strictStr
 in
   trivial (List.null pat || List.null search) $ all verify matches

prop_testRegressions :: Regressions -> Bool
prop_testRegressions r =
  let (pat, srch, idx) = regressionsToLazy r
      matches = kmpMatchLL pat srch 
  in
    if null matches
      then idx == -1
      else (head matches) == idx  
    
allTests = do
 quickCheck prop_testZero
 quickCheck prop_testSubset
 quickCheck prop_testCorrectPrefix
 quickCheck prop_testBadPat
 quickCheck prop_testMatch
 quickCheck prop_testEndPattern
 quickCheck prop_testRegressions
