{- OPTIONS_GHC -fbang-patterns -O2 #-}
{-
 Module      : Data.Sequence.KMP
 Copyright   : (c) Justin Bailey 2007 (jgbailey@gmail.com)
               (c) Chris Kuklewicz 2007 (haskell@list.mightyreason.com)
               (c) Daniel Fischer  (daniel.is.fischer@web.de)
 License     : BSD-style

 Maintainer  : jgbailey@gmail.com
 Stability   : experimental

 Implements Knuth-Morris-Pratt fast string matching for Data.Sequences.
 Original code by Justin Bailey, with significant
 enhancemants and optimizations by Chris Kuklewicz. Further enhancements to
 the computeLookup function provided by Daniel Fischer.

 Tests can be run by loading the module and evaluating "allTests".
-}
module Data.Sequence.KMP (kmpMatch)

where

import qualified Data.Array.Base as Base (unsafeAt, unsafeWrite, unsafeRead)
import Data.Array.Unboxed (UArray)
import Data.Array.IArray ((!), Array, IArray)
import qualified Data.Array.IArray as IArray (listArray, bounds)
import qualified Data.Array.ST as ST (runSTArray, runSTUArray, newArray_, newListArray, MArray, STUArray)
import qualified Data.IntMap as IntMap
import qualified Control.Monad.ST as State (ST)
import qualified Data.List as List (isPrefixOf, inits, isInfixOf)
import Data.Sequence (Seq, ViewL((:<), EmptyL))
import qualified Data.Sequence as Seq
import qualified Data.Foldable as F
import qualified Data.Traversable as T
import GHC.Int (Int64)

import Test.QuickCheck
import Debug.Trace (trace)

{-|
  Returns list of indices for a given substring in a search string, or the empty
  list if none were found.
  
  Uses the Knuth-Morris-Pratt fast string matching algorithm.
-}  
{-# INLINE kmpMatch #-}
kmpMatch :: (Eq a) => Seq a-- ^ Pattern to search for.
  -> Seq a -- ^ String to search.
  -> [Int] -- ^ List of indices where string was found.
kmpMatch pat = let search = kmpMatchS pat in search 

kmpMatchS :: (Eq a) =>  Seq a -> Seq a -> [Int]
kmpMatchS pat str =
  let !patLen = {-# SCC "_patLen" #-} Seq.length pat 
      !strLen = {-# SCC "_strLen" #-} Seq.length str  
      !patArray = {-# SCC "_patMap" #-} patToArray pat
      !lookupTable = {-# SCC "_computeLookup'" #-} computeLookup patArray -- lower bound of UArray must be 0 for Base.unsafeAt, but index 0 will never be looked up
      findMatch EmptyL !strIndex !patIndex
        | patIndex == patLen = {-# SCC "_findMatch_found_empty" #-}[strIndex - patLen]
        | otherwise = {-# SCC "_findMatch_empty_end" #-} []
      findMatch seqV@(e :< rest) !strIndex !patIndex
        | patIndex == patLen = {-# SCC "_findMatch_found" #-} (strIndex - patLen) : findMatch seqV strIndex 0 -- Success
        | strIndex == strLen ={-# SCC "_findMatch_end" #-} []
        | otherwise =
            if {-# SCC "_findMatch_eq" #-} (e == ({-# SCC "_patMapIndex" #-} Base.unsafeAt patArray patIndex))
              then {-# SCC "_findMatch_rest" #-} findMatch (Seq.viewl rest) (succ strIndex) (succ patIndex)
              else if patIndex == 0
                     then {-# SCC "_findMatch_restart" #-} findMatch (Seq.viewl rest) (succ strIndex) 0
                     else {-# SCC "_findMatch_shifted" #-}findMatch seqV strIndex (Base.unsafeAt lookupTable patIndex) -- here 1 <= patIndex <= patLen-1
  in
    if patLen == 0
      then [0]
      else findMatch (Seq.viewl str) 0 0

patToArray :: Seq a -> Array Int a
patToArray patSeq =
  ST.runSTArray 
    (do
      ar <- ST.newListArray (0,patLen) patList
      return ar)
  where
    !patLen = Seq.length patSeq
    !patList = F.foldr (\e l -> e : l) [] patSeq    
    
{- |
 Given a string representing a search pattern, this function
 returns a function which represents, for each prefix of that
 pattern, the maximally long prefix of the pattern which is a suffix
 of the indicated pattern segment.

 If there is no such prefix, 0 is returned. For the very first
 element, -1 is always returned.
 -}
computeLookup :: Eq a => Array Int a -> UArray Int Int
computeLookup pat
   = {-# SCC "_runSTUArray" #-} ST.runSTUArray
   (do ar <- ST.newArray_ (0,patLen)
       Base.unsafeWrite ar 0 (-1)
       Base.unsafeWrite ar 1 0
       let fill i
               | i > patLen = return ar
               | otherwise  =
                do let w = Base.unsafeAt pat (i-1)
                       help k
                           | k < 0 || w == Base.unsafeAt pat k =
                                do Base.unsafeWrite ar i (succ k)
                                   fill (i+1)
                           | otherwise = do m <- Base.unsafeRead ar k
                                            help m
                   n <- Base.unsafeRead ar (i-1)
                   help n
       {-# SCC "_fill" #-} fill 2)
   where
     !patLen = let (low, high) = IArray.bounds pat in high - low

     
-- Types, instances and utility functions for testing purposes.

newtype PatternChar = PatternChar Char
 deriving Show

instance Arbitrary PatternChar where
 arbitrary = oneof (map (return . PatternChar) ['a', 'b', 'c', 'd'])
 coarbitrary = undefined

-- Holds the search pattern, the search string, and the expected
-- position. Used to test for strings that have failed in the past.
data Regressions = R String String Int
  deriving Show
  
instance Arbitrary Regressions where
  arbitrary =
    oneof $ map return 
      [R "ccb" "abcdcccb" 5
       , R "bbaa" "bdcdbbbbaa" 6
       , R "cccadadc" "adaccccadadc" 4
       , R "bbdbbdaabdb" "dbbdbbbdbbdaabdb" 5
       , R "bbdbbd" "dbbdbbbdbbd" 5
       , R "ccbb" "acccbb" 2
       , R "bbbb" "dddcaaaddaaabdacbcccabbada" (-1) -- This string has caused an infinite loop.
       , R "bbc" "bbbbc" 2
       , R "a" "" (-1)
       , R "" "a" 0
       , R "" "" 0
       ]
  coarbitrary = undefined
    
  
patternsToString :: [PatternChar] -> String
patternsToString chars = foldr (\(PatternChar char) str -> char : str) [] chars

patternsToSequence :: [PatternChar] -> Seq Char
patternsToSequence chars = Seq.fromList $ foldr (\(PatternChar char) str -> char : str) [] chars

stringToSeq :: String -> Seq Char
stringToSeq str = Seq.fromList str

regressionsToSequence :: Regressions -> (Seq Char, Seq Char, Int)
regressionsToSequence (R pat srch idx) = (stringToSeq pat, stringToSeq srch, idx)

-- Test that 0 and 1 element always return 0, if present.
prop_testZero :: [PatternChar] -> Property
prop_testZero pat =
 let table = computeLookup (patToArray $ patternsToSequence pat)
 in
   trivial (null pat) $
   if length pat > 1
    then Base.unsafeAt table 0 == -1 && Base.unsafeAt table 1 == 0
    else Base.unsafeAt table 0 == -1

-- Test that all overlaps found are actually prefixes of the
-- pattern string
prop_testSubset :: [PatternChar] -> Property
prop_testSubset pat =
 let patStr = patternsToString pat
     table = computeLookup (patToArray $ patternsToSequence pat)
     testPrefix len =
       if Base.unsafeAt table len == 0
         then True
         else (take (Base.unsafeAt table len) patStr) `List.isPrefixOf` (take len patStr)
 in
   trivial (null pat) $
    all testPrefix [0 .. length pat - 1]


-- Test that the prefix given is the maximal prefix. That is,
-- add one more character makes it either equal to the string
-- or not a prefix.
prop_testCorrectPrefix :: [PatternChar] -> Property
prop_testCorrectPrefix pat =
 let patStr = patternsToString pat
     table = computeLookup (patToArray $ patternsToSequence pat)
     isNeverSuffix len =
       let origPrefix = take len patStr
           -- Drop 1 to remove empty list
           allPrefixes = drop 1 $ List.inits origPrefix
       in
         all (\p -> null p || p == origPrefix || not ((reverse p) `List.isPrefixOf` (reverse origPrefix))) allPrefixes
     -- True if the prefix returned from table for the length given produces
     -- a string which is a suffix of the original prefix.
     isRealSuffix len = (reverse (take (Base.unsafeAt table len) patStr)) `List.isPrefixOf` (reverse $ take len patStr)
     isLongestSuffix len =
       let prefixPlus = take (Base.unsafeAt table len + 1) patStr
           inputPrefix = take len patStr
       in
         prefixPlus == inputPrefix ||
           not ((reverse prefixPlus) `List.isPrefixOf` (reverse inputPrefix))
     testTable len =
       if Base.unsafeAt table len == 0
         then isNeverSuffix len
         else isRealSuffix len &&
               isLongestSuffix len
 in
   trivial (null pat) $
    all testTable [0 .. length pat - 1]


-- Verify that, if a match is found, it is where it's supposed to be in
-- the string and it can be independently found by other means.
prop_testMatch :: [PatternChar] -> [PatternChar] -> Property
prop_testMatch pat search =
 let patStr = patternsToString pat
     searchStr = patternsToString search
     strSeq = patternsToSequence search
     patSeq = patternsToSequence pat
     patLen = length patStr
     searchLen = length searchStr
     matches = kmpMatch patSeq strSeq
     verify matchIdx =
       if matchIdx > -1
         then (take patLen $ drop matchIdx $ searchStr) == patStr && patStr `List.isInfixOf` searchStr
         else (null patStr && null searchStr) || not (patStr `List.isInfixOf` searchStr)
 in
   trivial (null pat || null search) $
   classify (patLen > searchLen) "Bigger pattern than search" $
   classify (patLen < searchLen) "Smaller pattern than search" $
   classify (patLen == searchLen) "Equal pattern and search" $
   classify (not (null matches)) "Match Exists" $
   classify (null matches) "Match Doesn't Exist" $
    if patLen == 0
      then (head matches) == 0
      else all verify matches


-- Test a pattern that was known to fail.
prop_testBadPat :: [PatternChar] -> Property
prop_testBadPat search =
 let patSeq = stringToSeq "bbc"
     patLen = length "bbc"
     searchStr = patternsToString search
     strSeq = patternsToSequence search
     matches = kmpMatchS patSeq strSeq
     verify matchIdx =
            if matchIdx > -1
              then (take patLen $ drop matchIdx $ searchStr) == "bbc" && "bbc" `List.isInfixOf` searchStr
              else null searchStr || not ("bbc" `List.isInfixOf` searchStr)
 in
   trivial (null search) $
     if patLen == 0
      then (head matches) == 0
      else all verify matches

-- Test that a pattern on the end of the string is found OK.
prop_testEndPattern :: [PatternChar] -> [PatternChar] -> Property
prop_testEndPattern pat search =
 let patSeq = patternsToSequence pat
     searchSeq = patternsToSequence (search ++ pat)
     matches = kmpMatchS patSeq searchSeq
     searchStr = patternsToString (search ++ pat)
     patStr = patternsToString pat
     patLen = length patStr
     verify matchIdx = (take patLen $ drop matchIdx $ searchStr) == patStr && patStr `List.isInfixOf` searchStr
 in
   trivial (null pat || null search) $
    if patLen == 0
      then (head matches) == 0
      else all verify matches 


prop_testRegressions :: Regressions -> Bool
prop_testRegressions r =
  let (pat, srch, idx) = regressionsToSequence r
      matches = kmpMatchS pat srch 
  in
    if null matches
      then idx == -1
      else (head matches) == idx  
    
allTests = do
 quickCheck prop_testZero
 quickCheck prop_testSubset
 quickCheck prop_testCorrectPrefix
 quickCheck prop_testBadPat
 quickCheck prop_testMatch
 quickCheck prop_testEndPattern
 quickCheck prop_testRegressions